package info.bukova.isspst.services;

import static ch.lambdaj.Lambda.filter;
import info.bukova.isspst.Constants;
import info.bukova.isspst.Module;
import info.bukova.isspst.SessionData;
import info.bukova.isspst.dao.BaseDao;
import info.bukova.isspst.dao.QueryDao;
import info.bukova.isspst.data.DataModel;
import info.bukova.isspst.data.NumberSeries;
import info.bukova.isspst.filters.Filter;
import info.bukova.isspst.reporting.Report;
import info.bukova.isspst.services.numberseries.NumberSeriesService;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Set;

import javax.validation.ConstraintViolation;
import javax.validation.Validator;

import org.hibernate.NonUniqueResultException;
import org.hibernate.Query;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

public abstract class AbstractService<T extends DataModel> implements Service<T> {

	protected BaseDao<T> dao;
	private Validator validator;
	@Autowired
	protected SessionData sessionData;
	@Autowired
	protected QueryDao queryDao;

	private NumberSeriesService numberSeriesService;
	

	public NumberSeriesService getNumberSeriesService()
	{
		return numberSeriesService;
	}

	public void setNumberSeriesService(NumberSeriesService numberSeriesService)
	{
		this.numberSeriesService = numberSeriesService;
	}

	public void setDao(BaseDao<T> dao) {
		this.dao = dao;
	}
	
	public String getUpdateEntityPermission() {
		return "";
	}
	
	@Override
	@PreAuthorize("hasPermission(this, 'PERM_ADD')")
	public final T create() {
		return createEntity();
	}
	
	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_ADD')")
	public void add(T entity) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		validate(entity);
		entity.setCreated(new Date());
		dao.add(entity);
	}
	
	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_EDIT') or hasPermission(#entity, this.getUpdateEntityPermission())")
	public void update(T entity) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		validate(entity);
		entity.setModified(new Date());
		dao.modify(entity);
	}

	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_DELETE')")
	public void delete(T entity) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		dao.delete(entity);
	}
	
	@Override
	public void validate(T entity) {
		if (validator == null) {
			entity.setValid(true);
			return;
		}
		
		Set<ConstraintViolation<T>> violations = validator.validate(entity);
		
		if (!violations.isEmpty()) {
			ValidationException ex = new ValidationException();
			for(ConstraintViolation<T> v : violations) {
				String errKye = v.getPropertyPath().toString();
				ex.addMessage(errKye.isEmpty() ? "CLASS_ERR" : errKye, v.getMessage());
			}
			
			throw ex;
		}
		
		entity.setValid(true);
	}
	
	@Override
	public List<T> filterList(List<T> sourceList, Filter<T> filter) {
		return filter(filter.matcher(), sourceList);
	}

	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_READ')")
	public T getById(int id) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		return dao.getById(id);
	}
	
	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_READ')")
	public List<T> getAll() {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		return dao.getAll();
	}

	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_READ')")
	public List<T> execQuery(String query) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		return dao.execQuery(query);
	}

	@SuppressWarnings("unchecked")
	@Override
	@Transactional
	@PreAuthorize("hasPermission(this, 'PERM_READ')")
	public T selectSingle(String query) {
		if (dao == null) {
			throw new IsspstException("DAO is null");
		}
		
		try {
			Query q = dao.getQuery(query);
			return (T) q.uniqueResult();
		} catch (NonUniqueResultException e) {
			return null;
		}
	}

	public void setValidator(Validator validator) {
		this.validator = validator;
	}
	
	protected T createEntity() {
		return null;
	}
	
	@Override
	public Module getModule() {
		for (Module m : Constants.MODULES) {
			if (Arrays.asList(this.getClass().getInterfaces()).contains(m.getServiceClass())) {
				return m;
			}
		}
		
		return null;
	}
	
	public List<Report> getReports() {
		if (getModule() != null) {
			return getModule().getReports();
		} else {
			return null;
		}
	}

	protected String getNumberSerie()
	{
		String currentNumber = "";
		
		if (numberSeriesService != null)
		{
			String moduleName = this.getModule().getId();
			NumberSeries ns = numberSeriesService.getNumberSerie(moduleName);
			
			if (ns != null)
			{
				currentNumber = ns.getCurrentNumber();
				numberSeriesService.increase(ns);
			}
		}
		
		return currentNumber;
	}

	@Override
	@Transactional
	public void loadLazyData(T entity) {
		loadLazyData("", entity);
	}

	@Override
	@Transactional
	public void loadLazyData(String group, T entity) {
		Method[] methods = this.getClass().getMethods();
		
		for (Method m : methods) {
			Annotation[] annotations = m.getAnnotations();
			for (Annotation a : annotations) {
				if (a instanceof LazyLoader) {
					String[] values = ((LazyLoader) a).value();
					if (group.isEmpty() || values.length == 0 || Arrays.asList(values).contains(group)) {
						try {
							m.invoke(this, entity);
						} catch (IllegalArgumentException e) {
							throw new LazyLoadException(e.getCause());
						} catch (IllegalAccessException e) {
							throw new LazyLoadException(e.getCause());
						} catch (InvocationTargetException e) {
							throw new LazyLoadException(e.getCause());
						}
					}
				}
			}
		}
	}
}
